/* AA树的插入和删除函数
    代码选自网站en.wikipedia.org/wiki/AA_tree。
   要点：
     根据AA树的特征，在插入和删除一个数据后，重新调整树的结构，使之重新成为AA树
   关键字：自定义谓词，引理，引理依赖，幽灵变量
   说明:
     由于空树没有节点，自然没有level域，这不便于描述连空树也要考虑在内的节点层次值。
     所以，这里使用自定义谓词balance来解决这个问题。
*/
#include <stdlib.h>

#define MAX_LEVEL 1000

typedef struct node {
    int d; 
    int level; 
    struct node * l; 
    struct node * r;
} Node;
/*@ shape l, r : tree;*/

/*@
inductive Gt(int x, Node * p) = p == \null || x > p->d && Gt(x, p->l) && Gt(x, p->r);

inductive Lt(int x, Node * p) = p == \null || x < p->d && Lt(x, p->l) && Lt(x, p->r);

inductive BST(Node * p) = p == \null || BST(p->l) && BST(p->r) && Gt(p->d, p->l) && Lt(p->d, p->r);

inductive balance(Node * p, int m) =		// 当树非空时，根节点level域的值就是m。
    p == \null && m == 0 
    || p != \null && m > 0 && p->level == m && balance(p->l, m-1) && balance(p->r, m-1) 
    || p != \null && m > 0 && p->r != \null && p->level == m && p->r->level == m 
    && balance(p->l, m-1) && balance(p->r->l, m-1) && balance(p->r->r, m-1);

lemma Lt_lemma(both): \forall int x, int y. \forall Node * p. x < y && Lt(y, p) ==> Lt(x, p);

lemma Gt_lemma(both): \forall int x, int y. \forall Node * p. x > y && Gt(y, p) ==> Gt(x, p);
*/

//@ logic Node *oldt;

/*@
requires
    \tree(t) && BST(t) && Lt(y, t) && Gt(z, t) && oldt == t 
    && t->level == i && balance(t->l, i)
    && (t->l->r == \null && j == 0 
        || t->l->r->level == j && (t->l->r->r == \null && k == 0 || t->l->r->r->level == k))
     && balance(t->r, l);
assigns *oldt;
ensures
    \tree(\result->l) && oldt == \result->r && \tree(oldt)
    && BST(\result) && Lt(y, \result) && Gt(z, \result) 
    && \result->level == \result->r->level == i && balance(\result->l, i - 1) 
    && (j == 0 && \result->r->l == \null
        || j > 0 && balance(\result->r->l, j) && balance(\result->r->l->r, k))
    && balance(\result->r->r, l);
*/
Node * skew(Node * t) /*@ ghost (int const y, int const z, int const i, int const j, int const k, int const l) */;// {
//     Node *p;
//     /* swap the pointers of horizontal left links */
//     p = t->l;
//     t->l = p->r; 
//     p->r = t; 
//     t = p;
//     return t;
// }

/*@
requires 
    \tree(t->l) && \tree(t->r->l) && tmp == t->r->r && \tree(tmp) 
    && BST(t) && Lt(y, t) && Gt(z, t) && oldt == t
    && t->level == t->r->level == t->r->r->level == i <= MAX_LEVEL
    && balance(t->l, i - 1) && balance(t->r->l, i - 1) && balance(t->r->r->l, i - 1)
    && (t->r->r->r == \null && j == 0 
        || t->r->r->r->r == \null && t->r->r->r->level == j && k == 0 && balance(t->r->r->r->l, j - 1)
        || t->r->r->r->level == j && balance(t->r->r->r->l, j - 1) && balance(t->r->r->r->r, k) && balance(t->r->r->r->r->r, l));
assigns *oldt;
ensures	
    oldt == \result->l && \tree(oldt) && tmp == \result->r && \tree(tmp)
    && BST(\result) && Lt(y, \result) && Gt(z, \result)
    && \result->level == i + 1 && balance(\result->l, i) && balance(\result->l->r, i - 1) && \result->r->level == i && balance(\result->r->l, i - 1)
    && (j == 0 && \result->r->r == \null 
        || \result->r->r->r == \null && \result->r->r->level == j && k == 0 && balance(\result->r->r->l, j - 1)
        || \result->r->r->level == j && balance(\result->r->r->l, j - 1) && balance(\result->r->r->r, k) && balance(\result->r->r->r->r, l));
*/
Node * split(Node * t) /*@ ghost (int const y, int const z, int const i, int const j, int const k, int const l, Node * const tmp) */ ;//{
//     Node *p;
//     /* We have two horizontal right links. Take the middle node, elevate it, and return it. */
//     p = t->r; 
//     t->r = p->l; 
//     p->l = t;
//     p->level = p->level + 1;
//     t = p;
//     return t;
// }

/*@
requires t == \null && y < data < z && oldt == t && m == 0
    || t != \null && BST(t) && Lt(y, t) && Gt(z, t) && y < data < z && m <= MAX_LEVEL
        && oldt == t && balance(t, m) && balance(t->r, n);
assigns *oldt;
exits \exit_status == 1;
ensures 
    \result != \null && BST(\result) && Lt(y, \result) && Gt(z, \result) 
    && (m == 0 && oldt == \null && balance(\result, 1) && \result->r == \null
        || m == n > 0 && balance(\result->r, m)
            && (oldt == \result && (balance(\result, m) || balance(\result, m + 1))
                || oldt == \result->l && \tree(oldt) && \tree(\result->r) && balance(\result, m + 1))
        || m == n + 1 > 0 && balance(\result, m)
            && (oldt == \result || \tree(\result->l) && oldt == \result->r && \tree(oldt)));
*/
Node * AAinsert(Node * t, const int data) /*@ ghost (int const y, int const z, int const m, int const n) */ {
    //@ ghost int tmp1, tmp2, tmp3;
    if (t == NULL) {
        t = (Node*)malloc(sizeof(Node));
        if (t == NULL) { 
            exit(1);
        }
        t->l = NULL; 
        t->r = NULL;
        t->d = data; 
        t->level = 1;
    }
    else if (data < t->d) {
	    t->l = AAinsert(t->l, data)/*@ ghost (y, t->d, m - 1, (t->l == NULL || t->l->r == NULL) ? 0 : t->l->r->level) */;
    	if (t->l->level == t->level) {
            if (t->r != NULL && t->r->level == t->level) {
                t->level = t->level + 1;
            }
            else{
	            t = skew(t)/*@ ghost (y, z, m, m - 1, (t->l->r == NULL || t->l->r->r == NULL) ? 0 : t->l->r->r->level, m - 1) */;
            }
	    }
    }
    else if (data > t->d) {
        t->r = AAinsert(t->r, data)/*@ ghost (t->d, z, n, (t->r == NULL || t->r->r == NULL) ? 0 : t->r->r->level) */;
        if (t->r->r != NULL && t->r->r->level == t->level) {
            /*@ ghost
                tmp1 = 0;
                tmp2 = 0;
                tmp3 = 0;
                if (t->r->r->r != NULL) {
                    tmp1 = t->r->r->r->level;
                    if (t->r->r->r->r != NULL) {
                        tmp2 = t->r->r->r->r->level;
                        if (t->r->r->r->r->r != NULL) {
                            tmp3 = t->r->r->r->r->r->level;
                        }
                    }
                }
            */ 
            t = split(t)/*@ ghost (y, z, m, tmp1, tmp2, tmp3, t->r->r) */;
        }
    }
    return t;
}

/*@
inductive smallest(Node * p, int d) = 
    p->l == \null && d == p->d && Lt(d, p->r) 
    || p->l != \null && d < p->d && Lt(d, p->r) && smallest(p->l, d);

lemma smallest_gt_lemma: \forall int x, int y, Node * p. Gt(x, p) && smallest(p, y) ==> x > y;

lemma smallest_BST_lemma(relies on: Lt_lemma, smallest_gt_lemma): \forall int x, Node * p. BST(p) && smallest(p->l, x) ==> smallest(p, x);

lemma smallest_lt_lemma: \forall int x, int y, Node * p. smallest(p, x) && Lt(y, p) ==> x > y;
*/
/*@
requires t != \null && BST(t) && balance(t, m) && balance(t->r, n) && Lt(y, t) && Gt(z, t);
ensures t != \null && BST(t) && balance(t, m) && balance(t->r, n) && Lt(y, t) && Gt(z, t) && smallest(t, \result);
*/		
int minimal(Node * const t) /*@ ghost (int const y, int const z, int const m, int const n) */ {
    if (t->l == NULL) {
	    return t->d;
    }
    return minimal(t->l)/*@ ghost (y, t->d, m - 1, t->l->r == NULL ? 0 : t->l->r->level) */;
}

/*@
requires t == \null && oldt == t && m == 0
    || t != \null && BST(t) && oldt == t 
        && (Lt(y, t) && Gt(z, t) && y < data < z
            || smallest(t, y) && Gt(z, t) && y == data)
        && balance(t, m) && balance(t->r, n) && m <= MAX_LEVEL;
assigns oldt;
exits \exit_status == 1;
ensures \result == \null && oldt == \null && m == 0
    || \result == \null && \dangling(oldt) && m == 1
    || \result != \null && BST(\result) && Lt(y, \result) && Gt(z, \result) && m > 0
        && (m == n > 0 && balance(\result, m)
            && (oldt == \result->l || oldt == \result) && (balance(\result->r, m) || balance(\result->r, m - 1))
        || m == n + 1 > 0 && balance(\result->r, n)
            && ((oldt == \result || oldt == \result->r) && (balance(\result, m) || balance(\result, m - 1))
                || oldt == \result->l && balance(\result, m)));
*/
Node * AAdelete(Node * t, int const data) /*@ ghost (int const y, int const z, int const m, int const n) */ {
    int tmp;
    //@ ghost int tmp1, tmp2, tmp3;

    if (t != NULL) {
		if (data > t->d) {
		    t->r = AAdelete(t->r, data)/*@ ghost (t->d, z, n, 
            (t->r == NULL || t->r->r == NULL) ? 0 : t->r->r->level) */;
            /*@ assert BST(t) && Lt(y, t) && Gt(z, t) && t != \null && oldt == t
                && 0 < m <= MAX_LEVEL 
                && (   // branch E1, end
                    (m == n || m == n + 1) 
                    && balance(t, m) && balance(t->r, n)
                    ||    // branch E2, end
                    m == n && balance(t, m) && balance(t->r, m - 1)
                    ||    // branch R
                    m == n + 1 && t->level == m && balance(t->l, m - 1) && balance(t->r, m - 2)
                );
            */;
		}
		else if (data < t->d) {
		    t->l = AAdelete(t->l, data)/*@ ghost (y, t->d, m - 1, 
            (t->l == NULL || t->l->r == NULL) ? 0 : t->l->r->level) */;
            /*@ assert BST(t) && Lt(y, t) && Gt(z, t) && t != \null && oldt == t
                && 0 < m <= MAX_LEVEL 
                && (   // branch E1, end
                    (m == n || m == n + 1) 
                    && balance(t, m) && balance(t->r, n)
                    ||    // branch L1
                    m == n + 1 && t->level == m && balance(t->l, m - 2) && balance(t->r, m - 1)
                    ||    // branch L2
                    m == n == t->level == t->r->level && balance(t->l, m - 2) 
                    && balance(t->r->l, m - 1) && balance(t->r->r, m - 1)
                );
            */;
		}
		else if (t->r == NULL) {
			free(t);
            return NULL;
	    }
	    else {
            //@ ghost tmp1 = t->r->r == NULL ? 0 : t->r->r->level;
			t->d = minimal(t->r)/*@ ghost (t->d, z, n, tmp1) */;
			t->r = AAdelete(t->r, t->d)/*@ ghost (t->d, z, n, tmp1) */;
            /*@ assert BST(t) && Lt(y, t) && Gt(z, t) && t != \null && oldt == t
                && 0 < m <= MAX_LEVEL 
                && (   // branch E1, end
                    (m == n || m == n + 1) 
                    && balance(t, m) && balance(t->r, n)
                    ||    // branch E2, end
                    m == n && balance(t, m) && balance(t->r, m - 1)
                    ||    // branch R
                    m == n + 1 && t->level == m && balance(t->l, m - 1) && balance(t->r, m - 2)
                );
            */;
	    }
        /*@ assert BST(t) && Lt(y, t) && Gt(z, t) && t != \null && oldt == t
            && 0 < m <= MAX_LEVEL 
            && (   // branch E1, end
                (m == n || m == n + 1) 
                && balance(t, m) && balance(t->r, n)
                ||    // branch E2, end
                m == n && balance(t, m) && balance(t->r, m - 1)
                ||    // branch L1
                m == n + 1 && t->level == m && balance(t->l, m - 2) && balance(t->r, m - 1)
                ||    // branch L2
                m == n == t->level == t->r->level && balance(t->l, m - 2) 
                && balance(t->r->l, m - 1) && balance(t->r->r, m - 1)
                ||    // branch R
                m == n + 1 && t->level == m && balance(t->l, m - 1) && balance(t->r, m - 2)
            );
        */;
		/* Rebalance the tree. Decrease the level of all nodes in this level if necessary,
		   and then skew and split all nodes in the new level. */
        /* Decrease the level of all nodes in this level if necessary */
        if (t->l == NULL || t->r == NULL) {
            tmp = 1;
        }
        else if (t->l->level < t->r->level) {
            tmp = t->l->level + 1;
        }
        else {
            tmp = t->r->level + 1;
        }
        if (tmp < t->level) {
            t->level = tmp;
            if (t->r != NULL && tmp < t->r->level) {
                t->r->level = tmp;
            }
        }

        /* Skew and split all nodes in the new level */
        if (t->l != NULL && t->l->level == t->level) {
            // branch R
            /*@ ghost
                tmp1 = 0;
                tmp2 = 0;
                if (t->l->r != NULL) {
                    tmp1 = t->l->r->level;
                    if (t->l->r->r != NULL) {
                        tmp2 = t->l->r->r->level;
                    }
                }
            */ 
            t = skew(t)/*@ ghost (y, z, m - 1, tmp1, tmp2, m - 2) */;
        }

        if (t->r != NULL && t->r->l != NULL && t->r->l->level == t->r->level) {
            // branch L2, branch R.K2
            /*@ ghost
                tmp1 = 0;
                tmp2 = 0;
                tmp3 = 0;
                if (t->r->l->r != NULL) {
                    tmp1 = t->r->l->r->level;
                    if (t->r->l->r->r != NULL) {
                        tmp2 = t->r->l->r->r->level;
                    }
                }
                if (t->r->r != NULL) {
                    tmp3 = t->r->r->level;
                }
            */ 
            t->r = skew(t->r)/*@ ghost (t->d, z, m - 1, tmp1, tmp2, tmp3) */;
        }

        if (t->r != NULL && t->r->r != NULL && t->r->r->l != NULL && t->r->r->l->level == t->r->r->level) {
            // branch L2.KR1
            t->r->r = skew(t->r->r)/*@ ghost (t->r->d, z, m - 1, m - 2, 
            (t->r->r->l->r == NULL || t->r->r->l->r->r == NULL) ? 0 : t->r->r->l->r->r->level, 
            m - 1) */;
        }

        if (t->r != NULL && t->r->r != NULL && t->r->r->level == t->level) {
            // branch L1.2, branch L2.KR1.KRR, branch L2.KR2, branch R.K2.KR 
            /*@ ghost
                tmp1 = 0;
                tmp2 = 0;
                tmp3 = 0;
                if (t->r->r->r != NULL) {
                    tmp1 = t->r->r->r->level;
                    if (t->r->r->r->r != NULL) {
                        tmp2 = t->r->r->r->r->level;
                        if (t->r->r->r->r->r != NULL) {
                            tmp3 = t->r->r->r->r->r->level;
                        }
                    }
                }
            */ 
            t = split(t)/*@ ghost (y, z, m - 1, tmp1, tmp2, tmp3, t->r->r) */;
        }

        if (t->r != NULL && t->r->r != NULL && t->r->r->r != NULL && t->r->r->r->level == t->r->level) {
            // branch L2.KR1.KRR.1, branch L2.KR2.2 
            /*@ ghost
                tmp1 = 0;
                tmp2 = 0;
                tmp3 = 0;
                if (t->r->r->r->r != NULL) {
                    tmp1 = t->r->r->r->r->level;
                    if (t->r->r->r->r->r != NULL) {
                        tmp2 = t->r->r->r->r->r->level;
                        if (t->r->r->r->r->r->r != NULL) {
                            tmp3 = t->r->r->r->r->r->r->level;
                        }
                    }
                }
            */ 
            t->r = split(t->r)/*@ ghost (t->d, z, m - 1, tmp1, tmp2, tmp3, t->r->r->r) */;
        }
    }
    return t;
}